from __future__ import annotations

import io
import logging
import shutil
import zipfile
from contextlib import contextmanager
from contextlib import ExitStack
from operator import attrgetter
from pathlib import Path
from tempfile import TemporaryFile
from typing import IO
from typing import Iterator
from typing import Mapping
from typing import NamedTuple
from typing import NewType
from typing import TYPE_CHECKING
from urllib.request import Request
from urllib.request import urlopen

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.utils import NormalizedName
from packaging.version import InvalidVersion
from packaging.version import Version

from barnhunt.installer import github
from barnhunt.installer.metadata import InvalidDistribution
from barnhunt.installer.metadata import metadata_from_distdir
from barnhunt.installer.metadata import metadata_from_distzip

if TYPE_CHECKING:
    from _typeshed import StrPath
else:
    StrPath = object


DownloadUrl = NewType("DownloadUrl", str)


log = logging.getLogger()


class NoSuchDistribution(KeyError):
    """Raised if no matching distribution is found."""


class InkexProject(NamedTuple):
    install_dir: str  # FIXME: "type"?
    gh_owner: str
    gh_repo: str


INKEX_PROJECTS: dict[NormalizedName, InkexProject] = {
    canonicalize_name(name): project
    for name, project in [
        ("inkex-bh", InkexProject("extensions", "barnhunt", "inkex-bh")),
        ("bh-symbols", InkexProject("symbols", "barnhunt", "bh-symbols")),
    ]
}


class InkexRequirement(Requirement):
    url: DownloadUrl | None

    def __init__(self, requirement: str):
        super().__init__(requirement)
        self._validate()

    @property
    def project(self) -> InkexProject:
        return INKEX_PROJECTS[canonicalize_name(self.name)]

    def _validate(self) -> None:
        if canonicalize_name(self.name) not in INKEX_PROJECTS:
            raise ValueError(f"unknown requirement name {self.name!r}")
        if self.extras:
            raise ValueError("requirement extras not supported")
        if self.marker:
            raise ValueError("requirement markers not supported")


DEFAULT_REQUIREMENTS = tuple(InkexRequirement(name) for name in INKEX_PROJECTS)


@contextmanager
def _copy_to_tmp(fp: io.BufferedIOBase) -> Iterator[IO[bytes]]:
    with TemporaryFile() as tmp:
        for chunk in iter(fp.read1, b""):
            tmp.write(chunk)
        tmp.flush()
        tmp.seek(0)
        yield tmp


@contextmanager
def open_zipfile(url: DownloadUrl) -> Iterator[zipfile.ZipFile]:
    resp = urlopen(
        Request(
            url,
            headers={
                "User-Agent": "barnhunt (https://github.com/barnhunt/barnhunt)",
                "Accept": "application/zip, application/octet-stream;q=0.9, */*;q=0.1",
            },
        )
    )

    with ExitStack() as stack:
        resp = stack.enter_context(resp)
        if resp.seekable():
            fp = resp
        else:
            fp = stack.enter_context(_copy_to_tmp(resp))
        zf = zipfile.ZipFile(fp)
        yield stack.enter_context(zf)


class Installer:
    def __init__(
        self, target: StrPath, dry_run: bool = False, github_token: str | None = None
    ):
        self.target = Path(target)
        self.dry_run = dry_run
        self.github_token = github_token

    def install(
        self,
        requirement: InkexRequirement,
        pre_flag: bool = False,
        upgrade: bool = False,
    ) -> None:
        display_name = requirement.name
        canonical_name = canonicalize_name(requirement.name)
        specifiers = requirement.specifier
        project = requirement.project

        install_path = self.target / project.install_dir
        log.info("install_path is %s", install_path)
        installed = find_installed(install_path, canonical_name)
        versions = installed.keys()
        if len(versions) > 1:
            log.warning(
                "found multiple installed versions of %s: %s",
                display_name,
                ", ".join(map(str, versions)),
            )
        matching_version = max(
            specifiers.filter(versions, prereleases=True), default=None
        )
        if matching_version:
            if not upgrade and requirement.url is None:
                log.warning(
                    "%s==%s is already installed", display_name, matching_version
                )
                return
            log.debug("found installed %s==%s", display_name, matching_version)

        if requirement.url is None:
            dist_urls = find_distributions(project, self.github_token)
            log.debug(
                "found distributions for %s: %s",
                display_name,
                ", ".join(map(str, dist_urls.keys())),
            )
            prereleases = True if pre_flag else None
            try:
                version = max(specifiers.filter(dist_urls.keys(), prereleases))
            except ValueError as exc:
                raise NoSuchDistribution(
                    f"no distribution found for {requirement}"
                ) from exc

            if matching_version == version:
                log.warning("%s==%s is up-to-date", display_name, version)
                return
            download_url = dist_urls[version]
        else:
            version = None
            download_url = requirement.url

        # FIXME: check hashes?
        with open_zipfile(download_url) as zf:
            metadata = metadata_from_distzip(zf)
            # FIXME: better exceptions
            if version is not None:
                assert metadata.version == version
            assert metadata.canonical_name == canonical_name

            self._do_uninstall(display_name, installed)

            log.warning("installing %s==%s", display_name, metadata.version)
            if not self.dry_run:
                zf.extractall(install_path)

    def uninstall(self, requirement: InkexRequirement) -> None:
        canonical_name = canonicalize_name(requirement.name)
        install_path = self.target / requirement.project.install_dir
        installed = find_installed(install_path, canonical_name)
        self._do_uninstall(requirement.name, installed)

    def _do_uninstall(self, name: str, installed: Mapping[Version, Path]) -> None:
        for version, distdir in installed.items():
            log.warning("uninstalling %s==%s", name, version)
            log.info("removing %s", distdir)
            if not self.dry_run:
                shutil.rmtree(distdir)


def find_installed(
    install_path: StrPath, canonical_name: NormalizedName
) -> dict[Version, Path]:
    install_path = Path(install_path)
    installed: dict[Version, Path] = {}
    if install_path.is_dir():
        for distdir in install_path.iterdir():
            try:
                metadata = metadata_from_distdir(distdir)
            except InvalidDistribution:
                continue
            if metadata.canonical_name == canonical_name:
                # FIXME: warn/fix on duplicate version?
                installed[metadata.version] = distdir
    return installed


def find_distributions(
    project: InkexProject, github_token: str | None = None
) -> dict[Version, DownloadUrl]:
    """Find zip files assets attached to the releases for a GitHub repository.

    This returns a dict mapping versions to download URLs.

    Releases whose ``tag_name``s are not parseable as PEP440 version identifiers,
    and those that do not have a zip-file among their assets are ignored.
    """
    dists = {}
    for release in github.iter_releases(
        project.gh_owner, project.gh_repo, github_token=github_token
    ):
        try:
            version = Version(release.tag_name)
        except InvalidVersion:
            continue  # FIXME: warn?
        try:
            dists[version] = _get_download_url(release)
        except ValueError:
            continue  # FIXME: warn?
    return dists


def _get_download_url(release: github.Release) -> DownloadUrl:
    for asset in sorted(release.assets, key=attrgetter("updated_at"), reverse=True):
        if asset.content_type == "application/zip":
            if asset.browser_download_url:
                return DownloadUrl(asset.browser_download_url)
    raise ValueError("No zip-file asset found")



from __future__ import annotations

import os
from io import BytesIO
from pathlib import Path
from typing import Iterable
from typing import Iterator

import pytest
from lxml import etree

from barnhunt.coursemaps import _hash_dev_ino
from barnhunt.coursemaps import CourseMaps
from barnhunt.coursemaps import iter_coursemaps
from barnhunt.coursemaps import TemplateRenderer
from barnhunt.inkscape import svg
from barnhunt.layerinfo import parse_flagged_layer_info
from testlib import svg_maker


def get_labels(layers: Iterable[svg.LayerElement]) -> Iterator[str]:
    return map(svg.layer_label, layers)


@pytest.fixture
def template_renderer() -> TemplateRenderer:
    return TemplateRenderer(parse_flagged_layer_info)


@pytest.mark.parametrize("hidden", [True, False])
def test_render_templates(template_renderer: TemplateRenderer, hidden: bool) -> None:
    repls = svg.NSMAP.copy()
    repls["flags"] = "[h] " if hidden else ""
    xml = (
        """<g xmlns="%(svg)s" xmlns:inkscape="%(inkscape)s"
                inkscape:groupmode="layer"
                inkscape:label="%(flags)sJunk">
               <tspan>{{ somevar }}</tspan>
              </g>"""
        % repls
    )
    tree = etree.parse(BytesIO(xml.encode("utf-8")))
    result = template_renderer(tree, {"somevar": "foo"})
    expected = "{{ somevar }}" if hidden else "foo"
    assert result.getroot()[0].text == expected
    assert tree.getroot()[0].text == "{{ somevar }}"


def test_render_templates_failure(
    template_renderer: TemplateRenderer, caplog: pytest.LogCaptureFixture
) -> None:
    xml = '<tspan xmlns="%(svg)s">{{ foo() }}</tspan>' % svg.NSMAP
    tree = etree.parse(BytesIO(xml.encode("utf-8")))
    result = template_renderer(tree, {})
    assert result.getroot().text == "{{ foo() }}"
    assert "'foo' is undefined" in caplog.text


def test_is_hidden(template_renderer: TemplateRenderer) -> None:
    text = svg_maker.text("text")
    tspan = text[0]
    svg_maker.layer("[h] Layer", children=[text])

    assert template_renderer._is_hidden(tspan)
    assert template_renderer._is_hidden(text)


def test_is_hidden_layer(template_renderer: TemplateRenderer) -> None:
    elem = svg_maker.layer("[h] Layer")
    assert template_renderer._is_hidden_layer(elem)


def test_is_hidden_layer_not_hidden(template_renderer: TemplateRenderer) -> None:
    elem = svg_maker.layer("[o] Layer")
    assert not template_renderer._is_hidden_layer(elem)


def test_get_local_context(template_renderer: TemplateRenderer) -> None:
    xml = (
        """<g xmlns="%(svg)s"
                xmlns:inkscape="%(inkscape)s"
                inkscape:groupmode="layer"
                inkscape:label="The Layer">
               <tspan>{{ layer.label }}</tspan>
             </g>"""
        % svg.NSMAP
    )
    tree = etree.parse(BytesIO(xml.encode("utf-8")))
    tspan = tree.getroot()[0]
    context = template_renderer._get_local_context(tspan, dict(foo="bar"))
    assert context["foo"] == "bar"
    assert context["layer"].label == "The Layer"


class TestCourseMaps:
    @pytest.fixture
    def coursemaps(self) -> CourseMaps:
        return CourseMaps(layer_info_parser=parse_flagged_layer_info)

    @pytest.fixture
    def dummy_tree(self) -> svg.ElementTree:
        layer = svg_maker.layer
        return svg_maker.tree(
            layer(
                "Test",
                children=[
                    layer("[!base] Not in base"),
                ],
            ),
            layer(
                "[o] Overlay 2",
                children=[
                    layer("[o] Overlay 2.1"),
                ],
            ),
            layer(
                "Child",
                children=[
                    layer(
                        "[o] Overlay 1",
                        children=[
                            layer("[o] Overlay 1.3"),
                            layer("[h] Hidden 1.2"),
                            layer("[o] Overlay 1.1"),
                        ],
                    ),
                ],
            ),
            layer(
                "[h] Hidden",
                children=[
                    layer("Hidden Child"),
                ],
            ),
        )

    def test_init_with_context(self) -> None:
        coursemaps = CourseMaps(
            layer_info_parser=parse_flagged_layer_info, context={"foo": "bar"}
        )
        assert coursemaps.context["foo"] == "bar"

    def test_call(self, coursemaps: CourseMaps) -> None:
        layer = svg_maker.layer
        tree = svg_maker.tree(
            layer("[h] Hidden", children=[layer("Hidden Child")]),
            layer("Layer", children=[layer("Child")]),
        )

        result = list(coursemaps(tree))
        assert len(result) == 1
        context, pruned = result[0]
        assert context == {"course": None, "overlay": None, "overlays": ()}
        root = pruned.getroot()
        assert len(root) == 1 and svg.is_layer(root[0])
        assert svg.layer_label(root[0]) == "Layer"
        assert len(root[0]) == 1 and svg.is_layer(root[0][0])
        assert svg.layer_label(root[0][0]) == "Child"
        assert len(root[0][0]) == 0

        for layer_ in svg.walk_layers(root):
            # FIXME: assert not is_hidden(layer_)
            assert "display" not in layer_.get("style", "")

    def test_call_includes_hidden_clone_source(self, coursemaps: CourseMaps) -> None:
        layer = svg_maker.layer
        tree = svg_maker.tree(
            layer("[h] Hidden", children=[svg_maker.group(id="source")]),
            layer("Layer", children=[svg_maker.use("#source")]),
        )

        result = list(coursemaps(tree))
        assert len(result) == 1
        context, pruned = result[0]
        assert context == {"course": None, "overlay": None, "overlays": ()}
        root = pruned.getroot()
        assert len(root) == 2 and all(svg.is_layer(child) for child in root)
        assert "display:none;" in root[0].get("style", "")
        assert root[1].get("style") is None

    def test_get_context_no_overlays(self, coursemaps: CourseMaps) -> None:
        path: list[svg.LayerElement] = []
        context = coursemaps._get_context(path)
        assert context["course"] is None
        assert context["overlay"] is None

    def test_get_context_one_overlay(self, coursemaps: CourseMaps) -> None:
        path = [svg_maker.layer("[o] Foo")]
        context = coursemaps._get_context(path)
        assert len(context["overlays"]) == 1
        assert context["course"].label == "Foo"
        assert context["overlay"] is None

    def test_iter_overlays(
        self, coursemaps: CourseMaps, dummy_tree: svg.ElementTree
    ) -> None:
        root = dummy_tree.getroot()
        result = list(coursemaps._iter_overlays(root))
        labels = [
            (list(get_labels(path)), set(get_labels(cruft))) for path, cruft in result
        ]
        assert labels == [
            (
                ["[o] Overlay 1", "[o] Overlay 1.1"],
                {"[h] Hidden", "[h] Hidden 1.2", "[o] Overlay 1.3", "[o] Overlay 2"},
            ),
            (
                ["[o] Overlay 1", "[o] Overlay 1.3"],
                {"[h] Hidden", "[o] Overlay 1.1", "[h] Hidden 1.2", "[o] Overlay 2"},
            ),
            (["[o] Overlay 2", "[o] Overlay 2.1"], {"[h] Hidden", "[o] Overlay 1"}),
        ]

    @pytest.mark.parametrize(
        "output_basename, exclusions",
        [
            ("base", {"[!base] Not in base"}),
            ("notbase", set()),
        ],
    )
    def test_find_exclusions(
        self,
        coursemaps: CourseMaps,
        dummy_tree: svg.ElementTree,
        output_basename: str,
        exclusions: set[str],
    ) -> None:
        root = dummy_tree.getroot()
        result = coursemaps._find_exclusions(output_basename, root)
        assert set(get_labels(result)) == exclusions

    def test_find_overlays(
        self, coursemaps: CourseMaps, dummy_tree: svg.ElementTree
    ) -> None:
        root = dummy_tree.getroot()
        overlays, cruft = coursemaps._find_overlays(root)
        assert list(get_labels(overlays)) == ["[o] Overlay 1", "[o] Overlay 2"]
        assert set(get_labels(cruft)) == {"[h] Hidden"}


def test_hash_dev_ino() -> None:
    with open(__file__, "rb") as srcfile:
        st = os.stat(__file__)
        dev_ino = st.st_dev, st.st_ino
        assert _hash_dev_ino(srcfile) == hash(dev_ino)


# XXX: more unit tests for iter_coursemaps are probably in order
def test_iter_coursemaps_warns_if_no_random_seed(
    drawing_svg: Path, caplog: pytest.LogCaptureFixture
) -> None:
    svgfiles = [drawing_svg.open("rb")]
    for _ in iter_coursemaps(svgfiles):
        pass
    assert "no random-seed" in caplog.text



from __future__ import annotations

from typing import Iterable
from typing import overload

from lxml import etree
from lxml.builder import ElementMaker

from barnhunt._compat import Literal
from barnhunt.inkscape import svg


@overload
def get_by_id(
    tree: svg.ElementTree, id: str, raising: Literal[False]
) -> svg.Element | None:
    ...


@overload
def get_by_id(
    tree: svg.ElementTree, id: str, raising: Literal[True] = ...
) -> svg.Element:
    ...


def get_by_id(
    tree: svg.ElementTree, id: str, raising: bool = True
) -> svg.Element | None:
    elems: list[svg.Element] = tree.xpath("//*[@id=$id]", id=id)
    if len(elems) == 1:
        return elems[0]
    elif len(elems) > 1:
        raise ValueError(f"Multiple elements found for id {id!r}")
    elif raising:
        raise KeyError(id)
    return None


def get_by_ids(tree: svg.ElementTree, ids: Iterable[str]) -> Iterable[svg.Element]:
    for id in ids:
        yield get_by_id(tree, id)


_nsmap: dict[str | None, str] = dict(svg.NSMAP.items())
_nsmap[None] = svg.NSMAP["svg"]

_e = ElementMaker(namespace=_nsmap[None], nsmap=_nsmap)


class SvgMaker:
    @staticmethod
    def tree(*elems: svg.Element) -> svg.ElementTree:
        return etree.ElementTree(_e.svg({"id": "root"}, *elems))

    @staticmethod
    def defs(*elems: svg.Element) -> svg.Element:
        return _e.defs(*elems)

    @staticmethod
    def layer(
        label: str,
        id: str | None = None,
        children: Iterable[svg.Element] | None = None,
        visible: bool = True,
    ) -> svg.LayerElement:
        layer = _e.g(
            {
                svg.INKSCAPE_GROUPMODE: "layer",
                svg.INKSCAPE_LABEL: label,
                "id": id or "".join(word.lower() for word in label.split()),
            }
        )
        if not visible:
            layer.set("style", "display: hidden;")
        if children is not None:
            layer.extend(children)
        assert svg.is_layer(layer)
        return layer

    @staticmethod
    def text(label: str, id: str | None = None) -> svg.Element:
        attrib = {"id": id or "".join(word.lower() for word in label.split())}
        return _e.text(_e.tspan(attrib, label))

    @staticmethod
    def use(href: str, id: str | None = None) -> svg.Element:
        attrib = {
            "{http://www.w3.org/1999/xlink}href": href,
        }
        if id is not None:
            attrib["id"] = id
        return _e.use(attrib)

    @staticmethod
    def group(
        id: str | None = None,
        children: Iterable[svg.Element] | None = None,
    ) -> svg.Element:
        g = _e.g()
        if id is not None:
            g.set("id", id)
        if children is not None:
            g.extend(children)
        return g


svg_maker = SvgMaker()



from __future__ import annotations

import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

import pytest

from barnhunt.inkscape.runner import dwim_old_inkscape
from barnhunt.inkscape.runner import inkscape_runner


@pytest.fixture
def inkscape1_executable(inkscape_executable: str) -> str:
    if dwim_old_inkscape(inkscape_executable):
        pytest.skip("test requires inkscape >= 1.0")
    return inkscape_executable


@dataclass
class DummyInkscapeCommand:
    cli_args: Sequence[str] = ()
    shell_mode_cmdline: str = ""


@pytest.fixture(autouse=True)
def rl_enable_horiz_scroll_mode(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Configure GNU readline to scroll long input lines.

    When this scrolling is enabled, it screws up pexpect's command-line matching.
    """
    inputrc = tmp_path / ".inputrc"
    inputrc.write_text("set horizontal-scroll-mode on\n")
    monkeypatch.setitem(os.environ, "INPUTRC", os.fspath(inputrc))


@pytest.mark.parametrize("long_cmdline", [False, True])
def test_shell_mode_inkscape_version(
    long_cmdline: bool,
    inkscape1_executable: str,
    caplog: pytest.LogCaptureFixture,
) -> None:
    caplog.set_level(logging.DEBUG)
    runner = inkscape_runner(shell_mode=True, executable=inkscape1_executable)
    cmdline = "inkscape-version"
    if long_cmdline:
        cmdline += f":{'x' * 256}"
    runner.run(DummyInkscapeCommand(shell_mode_cmdline=cmdline))

    warnings = [
        message
        for _, level, message in caplog.record_tuples
        if level >= logging.WARNING
    ]
    assert 1 <= len(warnings) <= 2

    unexpected_lines = [
        line[6:] for line in warnings[-1].splitlines() if line.startswith("!! => ")
    ]
    assert len(unexpected_lines) == 1
    assert re.match(r"\AInkscape \d+\.\d+\.\d+ ", unexpected_lines[0])



import dataclasses
import json
import sys
import zipfile
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar

import marshmallow
from marshmallow_dataclass import class_schema
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.utils import NormalizedName
from packaging.version import Version

if TYPE_CHECKING:
    from _typeshed import StrPath
else:
    StrPath = object

__all__ = [
    "InvalidDistribution",
    "Metadata",
    "metadata_from_distdir",
    "metadata_from_distzip",
    "metadata_from_json",
]


class InvalidDistribution(ValueError):
    """Raised if distribution zipfile is screwed up."""


@dataclass
class Metadata_2_1:
    # NB: API based on packaging.metadata.Metadata (unreleased as of packaging==21.3 but
    # in git)
    name: str
    version: Version
    if sys.version_info >= (3, 10):
        _: dataclasses.KW_ONLY
    summary: Optional[str] = None
    description: Optional[str] = None
    description_content_type: Optional[str] = None
    keywords: Optional[List[str]] = None
    license: Optional[str] = None
    home_page: Optional[str] = None
    download_url: Optional[str] = None
    project_urls: Optional[List[Tuple[str, str]]] = None
    author: Optional[str] = None
    author_email: Optional[str] = None
    maintainer: Optional[str] = None
    maintainer_email: Optional[str] = None
    classifiers: Optional[List[str]] = None

    platforms: Optional[List[str]] = None
    supported_platforms: Optional[List[str]] = None

    requires_python: SpecifierSet = field(default_factory=SpecifierSet)
    requires_dists: Optional[List[Requirement]] = None
    provides_extras: Optional[List[NormalizedName]] = None
    # dynamic_fields: Optional[List[DynamicField]]
    requires_externals: Optional[List[str]] = None

    provides_dists: Optional[List[str]] = None
    obsoletes_dists: Optional[List[str]] = None

    @property
    def display_name(self) -> str:
        return self.name

    @display_name.setter
    def display_name(self, value: str) -> None:
        self.name = value

    @property
    def canonical_name(self) -> NormalizedName:
        return canonicalize_name(self.name)


T = TypeVar("T")


class TypeField(marshmallow.fields.Field, Generic[T]):
    def _deserialize(self, value: Any, *args: Any, **kwargs: Any) -> T:
        T_ = self.__orig_class__.__args__[0]  # type: ignore[attr-defined]
        if isinstance(value, T_):
            return value  # type: ignore[no-any-return]
        try:
            return T_(value)  # type: ignore[no-any-return]
        except ValueError as exc:
            raise marshmallow.ValidationError(str(exc)) from exc


class NormalizedNameField(marshmallow.fields.Field, Generic[T]):
    def _deserialize(self, value: Any, *args: Any, **kwargs: Any) -> NormalizedName:
        return canonicalize_name(value)


class SchemaBase(marshmallow.Schema):
    TYPE_MAPPING = {
        NormalizedName: NormalizedNameField,
        Requirement: TypeField[Requirement],
        SpecifierSet: TypeField[SpecifierSet],
        Version: TypeField[Version],
    }

    class Meta:
        unknown = marshmallow.EXCLUDE


Metadata_2_1_Schema = class_schema(Metadata_2_1, base_schema=SchemaBase)

Metadata = Metadata_2_1


def metadata_from_json(data: Dict[str, Any]) -> Metadata:
    schema = Metadata_2_1_Schema()
    metadata: Metadata_2_1 = schema.load(data)
    return metadata


def metadata_from_distzip(distzip: zipfile.ZipFile) -> Metadata:
    # XXX: Could use zipfile.Path to clean this up in python > 3.8
    for info in distzip.infolist():
        # FIXME: check for multiple METADATA.json?
        if info.filename.endswith("/METADATA.json"):
            json_data = json.loads(distzip.read(info))
            return metadata_from_json(json_data)
    # FIXME: better error if METADATA.json not found
    raise InvalidDistribution("No METADATA.json file found in zip file")


def metadata_from_distdir(dir_path: StrPath) -> Metadata:
    mdfile = Path(dir_path, "METADATA.json")
    try:
        with mdfile.open("rb") as fp:
            json_data = json.load(fp)
    except (FileNotFoundError, NotADirectoryError, ValueError) as exc:
        raise InvalidDistribution("No METADATA.json file found in dist") from exc
    return metadata_from_json(json_data)



import logging
import os
import re
import sys
import threading
from dataclasses import dataclass
from subprocess import CalledProcessError
from typing import Any
from typing import Callable
from typing import Sequence
from typing import Type

import pytest

from barnhunt.inkscape.runner import CliRunner
from barnhunt.inkscape.runner import dwim_old_inkscape
from barnhunt.inkscape.runner import ExportPdfCommand
from barnhunt.inkscape.runner import ExportPdfCommand_0_9x
from barnhunt.inkscape.runner import ExportPdfCommand_1_0
from barnhunt.inkscape.runner import inkscape_runner
from barnhunt.inkscape.runner import InkscapeApi
from barnhunt.inkscape.runner import log_output
from barnhunt.inkscape.runner import Runner
from barnhunt.inkscape.runner import ShellModeRunner


class TestExportPdfCommand:
    @pytest.fixture(params=[ExportPdfCommand_0_9x, ExportPdfCommand_1_0])
    def command_class(self, request: pytest.FixtureRequest) -> ExportPdfCommand:
        return request.param  # type: ignore[no-any-return]

    def test_no_pdf_version(self, command_class: Type[ExportPdfCommand]) -> None:
        command = command_class("in.svg", "out.pdf")
        assert all("export-pdf-version" not in arg for arg in command.cli_args)
        assert "export-pdf-version" not in command.shell_mode_cmdline

    def test_pdf_version(self, command_class: Type[ExportPdfCommand]) -> None:
        command = command_class("in.svg", "out.pdf", "1.23")
        assert "--export-pdf-version=1.23" in command.cli_args
        if command_class is ExportPdfCommand_0_9x:
            assert " --export-pdf-version=1.23 " in command.shell_mode_cmdline
        else:
            assert "; export-pdf-version:1.23;" in command.shell_mode_cmdline


@dataclass
class DummyInkscapeCommand:
    cli_args: Sequence[str] = ()
    shell_mode_cmdline: str = ""


class TestCliRunner:
    @pytest.fixture
    def runner(self) -> CliRunner:
        api = InkscapeApi(
            export_pdf_command=lambda *_args: DummyInkscapeCommand(),
        )
        return CliRunner(api=api, executable=sys.executable)

    def test_success(self, runner: Runner) -> None:
        command = DummyInkscapeCommand(cli_args=("-c", ""))
        runner.run(command)

    def test_failure(self, runner: Runner) -> None:
        command = DummyInkscapeCommand(cli_args=("-c", "import sys; sys.exit(1)"))
        with pytest.raises(CalledProcessError):
            runner.run(command)

    def test_logs_output(
        self, runner: Runner, caplog: pytest.LogCaptureFixture
    ) -> None:
        command = DummyInkscapeCommand(cli_args=("-c", "print('foo')"))
        runner.run(command)
        assert "foo" in caplog.text

    def test_export_pdf(self, caplog: pytest.LogCaptureFixture) -> None:
        # mostly here for coverage
        command = DummyInkscapeCommand(cli_args=("-c", "print('foo')"))
        api = InkscapeApi(export_pdf_command=lambda *_args: command)
        runner = CliRunner(api=api, executable=sys.executable)
        runner.export_pdf("ignored", "ignored")
        assert "foo" in caplog.text

    def test_close(self, runner: Runner) -> None:
        # vanity test for coverage
        runner.close()


class TestShellModeRunner:
    @pytest.fixture
    def runner(self) -> ShellModeRunner:
        here = os.path.dirname(os.path.abspath(__file__))
        dummy_inkscape_py = os.path.join(here, "dummy_inkscape.py")
        api = InkscapeApi(
            export_pdf_command=lambda *_args: DummyInkscapeCommand(),
            shell_mode_cruft_patterns=(r"^DummyShellmodeInkscape$",),
        )
        return ShellModeRunner(
            api=api,
            executable=sys.executable,
            inkscape_args=(dummy_inkscape_py,),
            timeout=30,
        )

    def test_success(self, runner: Runner, caplog: pytest.LogCaptureFixture) -> None:
        caplog.set_level(logging.INFO)
        command = DummyInkscapeCommand(shell_mode_cmdline="true")
        runner.run(command)
        assert len(caplog.messages) == 0

    def test_logs_output(
        self, runner: Runner, caplog: pytest.LogCaptureFixture
    ) -> None:
        caplog.set_level(logging.INFO)
        command = DummyInkscapeCommand(shell_mode_cmdline="echo foo")
        runner.run(command)
        assert len(caplog.messages) == 1
        assert "foo" in caplog.messages[0]

    def test_close(self, runner: ShellModeRunner) -> None:
        command = DummyInkscapeCommand(shell_mode_cmdline="true")
        with runner:
            runner.run(command)
            proc = runner._proc
            assert proc is not None
            assert proc.poll() is None, "child process is alive"

        # check that child is killed within a reasonable time
        proc.wait(1.0)

    @pytest.mark.xfail(sys.platform == "darwin", reason="broken on macOS")
    def test_thread_safe(
        self, runner: ShellModeRunner, caplog: pytest.LogCaptureFixture
    ) -> None:
        nthreads = 16
        procs = set()
        command = DummyInkscapeCommand(shell_mode_cmdline="true")

        def target() -> None:
            runner.run(command)
            procs.add(runner._proc)

        caplog.set_level(logging.INFO)
        run_in_threads(target, nthreads=nthreads)
        assert len(caplog.messages) == 0
        assert len(procs) == nthreads
        for proc in procs:
            # Check that all subprocesses have exited
            assert proc is not None
            assert proc.wait(1.0) == 0

    def test_proc_is_none_if_not_started(self, runner: ShellModeRunner) -> None:
        assert runner._proc is None


def run_in_threads(target: Callable[[], Any], nthreads: int) -> None:
    threads = [threading.Thread(target=target) for _ in range(nthreads)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()


def test_log_output(caplog: pytest.LogCaptureFixture) -> None:
    log_output("foo")
    assert len(caplog.messages) == 1
    assert re.search(r"(?m)^!! => foo", caplog.text)


def test_log_output_squelches_cruft(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.INFO)
    log_output("foo", cruft_patterns=(r"^foo$",))
    assert len(caplog.messages) == 0


@pytest.mark.parametrize(
    ("shell_mode", "old_inkscape", "runner_class", "export_pdf_command"),
    [
        (False, False, CliRunner, ExportPdfCommand_1_0),
        (False, True, CliRunner, ExportPdfCommand_0_9x),
        (True, False, ShellModeRunner, ExportPdfCommand_1_0),
        (True, True, ShellModeRunner, ExportPdfCommand_0_9x),
    ],
)
def test_inkscape_runner(
    shell_mode: bool,
    old_inkscape: bool,
    runner_class: Type[Runner],
    export_pdf_command: Any,
) -> None:
    runner = inkscape_runner(shell_mode, old_inkscape=old_inkscape)
    assert isinstance(runner, runner_class)
    assert runner.api.export_pdf_command is export_pdf_command


def test_dwim_old_inkscape(inkscape_executable: str) -> None:
    assert isinstance(dwim_old_inkscape(inkscape_executable), bool)


def test_dwim_old_inkscape_warns_on_garbage(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.WARNING)
    python = sys.executable
    assert dwim_old_inkscape(python) is False
    assert "Can not determine Inkscape version" in caplog.text



""" Helpers to operator on Inkscape SVG files.

"""
from __future__ import annotations

import copy
import dataclasses
import math
import random
from collections import deque
from functools import lru_cache
from itertools import islice
from typing import cast
from typing import Collection
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import NewType
from typing import overload
from typing import TYPE_CHECKING

from lxml import etree

from .._compat import TypeGuard
from .css import InlineCSS

# XML element
Element = etree._Element
LayerElement = NewType("LayerElement", Element)

if TYPE_CHECKING:
    ElementTree = etree._ElementTree[Element]
else:
    ElementTree = object


NSMAP = {
    "svg": "http://www.w3.org/2000/svg",
    "inkscape": "http://www.inkscape.org/namespaces/inkscape",
    "xlink": "http://www.w3.org/1999/xlink",
    "bh": "http://dairiki.org/barnhunt/inkscape-extensions",
}

etree.register_namespace("bh", NSMAP["bh"])  # type: ignore[attr-defined]


def _qname(tag: str) -> str:
    prefix, sep, localname = tag.rpartition(":")
    assert sep
    return f"{{{NSMAP[prefix]}}}{localname}"


SVG_SVG_TAG = _qname("svg:svg")
SVG_G_TAG = _qname("svg:g")
SVG_TEXT_TAG = _qname("svg:text")
SVG_TSPAN_TAG = _qname("svg:tspan")
SVG_USE_TAG = _qname("svg:use")

INKSCAPE_GROUPMODE = _qname("inkscape:groupmode")
INKSCAPE_LABEL = _qname("inkscape:label")
XLINK_HREF = _qname("xlink:href")

LAYER_XP = f'{SVG_G_TAG}[@{INKSCAPE_GROUPMODE}="layer"]'

BH_RANDOM_SEED = _qname("bh:random-seed")


def walk_layers(elem: Element) -> Iterator[LayerElement]:
    """Iterate over all layers under elem.

    The layers are returned depth-first order, however at each level the
    layers are iterated over in reverse order.  (In the SVG tree, layers
    are listed from bottom to top in the stacking order.  We list them
    from top to bottom.)

    """
    for layer, _children in walk_layers2(elem):
        yield layer


def walk_layers2(elem: Element) -> Iterator[tuple[LayerElement, list[LayerElement]]]:
    """Iterate over all layers under elem.

    This is just like ``walk_layers``, except that it yields a
    sequence of ``(elem, children)`` pairs.  ``Children`` will be a
    list of the sub-layers of ``elem``.  It can be modified in-place
    to "prune" the traversal of the layer tree.

    """
    nodes = elem.findall("./" + LAYER_XP)
    while nodes:
        elem = LayerElement(nodes.pop())
        children = elem.findall("./" + LAYER_XP)
        children.reverse()
        yield elem, cast(List[LayerElement], children)
        nodes.extend(reversed(children))


def is_layer(elem: Element) -> TypeGuard[LayerElement]:
    """Is elem an Inkscape layer element?"""
    return elem.tag == SVG_G_TAG and elem.get(INKSCAPE_GROUPMODE) == "layer"


def lineage(elem: Element) -> Iterator[Element]:
    """Iterate over elem and its ancestors.

    The first element returned will be elem itself. Next comes elem's parent,
    then grandparent, and so on...

    """
    ancestor: Element | None = elem
    while ancestor is not None:
        yield ancestor
        ancestor = ancestor.getparent()


def parent_layer(elem: Element) -> LayerElement | None:
    """Find the layer which contains elem.

    Returns the element for the Inkscape layer which contains ``elem``.

    """
    return next(ancestor_layers(elem), None)


def ancestor_layers(elem: Element) -> Iterator[LayerElement]:
    """Iterate the ancestor layers of element.

    Yields, first, the layer that contains ``elem``, then the layer that
    contains that layer, and so on.

    """
    for parent in islice(lineage(elem), 1, None):
        if is_layer(parent):
            yield parent


def sibling_layers(elem: Element) -> Iterator[LayerElement]:
    """Iterate over sibling layers, *not* including self."""
    parent = elem.getparent()
    if parent is None:
        return
    for sibling in parent:
        if sibling is not elem and is_layer(sibling):
            yield sibling


def ensure_visible(elem: Element) -> None:
    style = InlineCSS(elem.get("style"))
    if style.get("display", "").strip() == "none":
        style["display"] = "inline"
        elem.set("style", style.serialize())


def set_hidden(elem: Element) -> None:
    style = InlineCSS(elem.get("style"))
    style["display"] = "none"
    elem.set("style", style.serialize())


def layer_label(layer: LayerElement) -> str:
    """Get the label of on Inkscape layer"""
    return layer.get(INKSCAPE_LABEL) or ""


class CloneInfo(NamedTuple):
    elem: Element
    ref: Element


@lru_cache(128)
def _compile_xpath(xpath: str) -> etree.XPath:
    return etree.XPath(xpath, namespaces=NSMAP)


def find_clones(tree: ElementTree) -> Iterator[CloneInfo]:
    """Find clones in document.

    “Clones” are <svg:use> elements whose href points to an element outside of the
    <svg:defs> element.

    Returns an iterable of (elem, ref) pairs where ``elem`` is the <svg:use> element,
    and ``ref`` is the element referenced by the ``href``.
    """
    # catalog elements outside of <svg:defs>
    find_elems_outside_defs = _compile_xpath("//*[@id and not(ancestor::svg:defs)]")
    elems_by_id = {elem.get("id"): elem for elem in find_elems_outside_defs(tree)}

    find_clones_outside_defs = _compile_xpath("//svg:use[not(ancestor::svg:defs)]")
    for elem in find_clones_outside_defs(tree):
        href = elem.get("href") or elem.get(XLINK_HREF)
        if href is not None and href.startswith("#"):  # XXX: worry about "url(#foo)"?
            ref = elems_by_id.get(href[1:])
            if ref is not None:
                yield CloneInfo(elem, ref)


def find_hidden_clone_source_layers(
    tree: ElementTree, hidden_layers: Iterable[LayerElement]
) -> set[LayerElement]:
    """Find hidden layers containing sources for visible “clones”."""
    hidden_source_layers: set[LayerElement] = set()
    omitted_layers = set(hidden_layers)
    clones: deque[CloneInfo] = deque(find_clones(tree))
    hidden_clones: list[CloneInfo] = []

    def is_visible(elem: Element) -> bool:
        return omitted_layers.isdisjoint(ancestor_layers(elem))

    while clones:
        clone = clones.popleft()
        if not is_visible(clone.elem):
            hidden_clones.append(clone)
        elif not is_visible(clone.ref):
            ref_layer = parent_layer(clone.ref)
            assert ref_layer is not None
            hidden_source_layers.add(ref_layer)
            omitted_layers.remove(ref_layer)
            # recheck all the clones that were not visible
            clones.extend(hidden_clones)
            hidden_clones.clear()

    return hidden_source_layers


def copy_etree(
    tree: ElementTree,
    omit_elements: Collection[Element] | None = None,
    update_nsmap: Mapping[str | None, str] | None = None,
) -> ElementTree:
    """Copy an entire element tree, possibly making modifications.

    Any elements listed in ``omit_elements`` (along with the
    descendants of any such elements) will be omitted entirely from
    the copy.

    The namespace map of the copied root element will be augmented
    with any mappings specified by ``update_nsmap``.

    """
    omit_elems = set(omit_elements or ())

    def copy_elem(
        elem: Element, nsmap: Mapping[str | None, str] | None = None
    ) -> Element:
        if nsmap is None and omit_elems.isdisjoint(elem.iter()):
            # No descendants are in omit_elements.
            return copy.deepcopy(elem)  # speed optimization

        rv = etree.Element(elem.tag, attrib=elem.attrib, nsmap=nsmap)
        rv.text = elem.text
        rv.tail = elem.tail
        rv.extend(copy_elem(child) for child in elem if child not in omit_elems)
        return rv

    root = tree.getroot()
    rv = copy.copy(tree)
    assert rv.getroot() is root
    nsmap = root.nsmap
    if update_nsmap is not None:
        nsmap.update(update_nsmap)
    rv._setroot(copy_elem(root, nsmap=nsmap))
    return rv


@dataclasses.dataclass
class EnsureId:
    """A helper to ensure that elements have id attributes.

    A instance of this class is a callable which may be applied to elements
    from ``tree``.  If the element has an id attribute set, it is returned;
    otherwise, a new unique id is assigned to the element and the new id is
    returned.

    """

    tree: ElementTree
    mindigits: int = 5
    sparedigits: int = 1
    _unique_ids: Iterator[str] | None = dataclasses.field(default=None, init=False)

    def __call__(self, elem: Element) -> str:
        id_ = elem.get("id")
        if not id_:
            if self._unique_ids is None:
                self._unique_ids = self._iter_unique_ids()
            id_ = next(self._unique_ids)
            elem.set("id", id_)
        return id_

    def _iter_unique_ids(self) -> Iterator[str]:
        seen = {elem.get("id") for elem in self.tree.iterfind("//*[@id]")}

        def random_id() -> str:
            return f"bh-{random.randint(1, 10 ** ndigits - 1):0{ndigits}d}"

        while True:
            mindigits = int(math.log10(max(len(seen), 1))) + 1
            ndigits = max(mindigits, self.mindigits) + self.sparedigits
            new_id = random_id()
            while new_id in seen:
                new_id = random_id()
            seen.add(new_id)
            yield new_id


def _svg_attrib(tree: ElementTree) -> etree._Attrib:
    svg_elem = tree.getroot()
    if svg_elem.tag != SVG_SVG_TAG:
        raise ValueError(f"Expected XML root to be an <svg> tag, not <{svg_elem.tag}>")
    return svg_elem.attrib


@overload
def get_svg_attrib(
    tree: ElementTree, attr: str | bytes | etree.QName, default: str
) -> str:
    ...


@overload
def get_svg_attrib(
    tree: ElementTree, attr: str | bytes | etree.QName, default: None = ...
) -> str | None:
    ...


def get_svg_attrib(
    tree: ElementTree, attr: str | bytes | etree.QName, default: str | None = None
) -> str | None:
    """Get XML attribute from root <svg> element.

    The attribute name, `attr`, should be namedspaced.

    Returns `default` (default `None`) if the attribute does not exist.

    """
    return _svg_attrib(tree).get(attr, default)


def set_svg_attrib(
    tree: ElementTree, attr: str | bytes | etree.QName, value: str
) -> None:
    """Get XML attribute on root <svg> element.

    The attribute specified by the namedspaced `attr` is set to `value`.

    `Tree` is modified *in place*.
    """
    _svg_attrib(tree)[attr] = value


@overload
def get_random_seed(tree: ElementTree, default: int) -> int:
    ...


@overload
def get_random_seed(tree: ElementTree, default: None = ...) -> int | None:
    ...


def get_random_seed(tree: ElementTree, default: int | None = None) -> int | None:
    value = get_svg_attrib(tree, BH_RANDOM_SEED)
    if value is None:
        return default
    try:
        return int(value, base=0)
    except ValueError as ex:
        raise ValueError(
            f"Expected integer, not {value!r} for /svg/@bh:random-seed"
        ) from ex


def set_random_seed(tree: ElementTree, value: int) -> None:
    set_svg_attrib(tree, BH_RANDOM_SEED, f"{value:d}")
